import java.util.List;

/**
 * @author Erwan LBP
 * @version 1.0
 * @since 22-02-2017
 */
public class BalasHammer {

    private int[][] tab_result;
    private int[][] tab_deb;
    private int nbL, nbC, colDemande, ligStock;
    private boolean[][] crossed;

    public BalasHammer(List<List<Integer>> listDeb) {
        this.ligStock = listDeb.size();
        this.colDemande = listDeb.get(0).size();
        this.nbL = ligStock - 1;
        this.nbC = colDemande - 1;

        this.tab_result = new int[nbL][nbC];
        for (int i = 0; i < nbL; i++)
            for (int j = 0; j < nbC; j++)
                tab_result[i][j] = 0;

        this.tab_deb = new int[ligStock][colDemande];
        for (int i = 0; i < ligStock; i++)
            for (int j = 0; j < colDemande; j++)
                tab_deb[i][j] = listDeb.get(i).get(j);

        crossed = new boolean[nbL][nbC];
        for (int i = 0; i < nbL; i++)
            for (int j = 0; j < nbC; j++)
                crossed[i][j] = false;
    }

    public boolean run() {
        while (!fini()) {
            int tmp;
            int indiceMaxDeltaLine = -1, maxDeltaLine = -1;
            int indiceMaxDeltaCol = -1, maxDeltaCol = -1;

            for (int i = 0; i < nbL; i++) {
                tmp = calculerDelta(tab_deb, i, true);
                if (tmp >= maxDeltaLine) {
                    maxDeltaLine = tmp;
                    indiceMaxDeltaLine = i;
                }
            }
            for (int i = 0; i < nbC; i++) {
                tmp = calculerDelta(tab_deb, i, false);
                if (tmp >= maxDeltaCol) {
                    maxDeltaCol = tmp;
                    indiceMaxDeltaCol = i;
                }
            }

            int indice2;
            if (maxDeltaLine > maxDeltaCol) {
                indice2 = calculerIndiceMin(tab_deb, indiceMaxDeltaLine, true);
                tab_result[indiceMaxDeltaLine][indice2] = Math.min(tab_deb[indiceMaxDeltaLine][colDemande - 1], tab_deb[ligStock - 1][indice2]);
                tab_deb[indiceMaxDeltaLine][colDemande - 1] -= tab_result[indiceMaxDeltaLine][indice2];
                tab_deb[ligStock - 1][indice2] -= tab_result[indiceMaxDeltaLine][indice2];
                if (tab_deb[indiceMaxDeltaLine][colDemande - 1] == 0)
                    cross(indiceMaxDeltaLine, true);
                if (tab_deb[ligStock - 1][indice2] == 0)
                    cross(indice2, false);
            } else {
                indice2 = calculerIndiceMin(tab_deb, indiceMaxDeltaCol, false);
                tab_result[indice2][indiceMaxDeltaCol] = Math.min(tab_deb[indice2][colDemande - 1], tab_deb[ligStock - 1][indiceMaxDeltaCol]);
                tab_deb[indice2][colDemande - 1] -= tab_result[indice2][indiceMaxDeltaCol];
                tab_deb[ligStock - 1][indiceMaxDeltaCol] -= tab_result[indice2][indiceMaxDeltaCol];
                if (tab_deb[ligStock - 1][indiceMaxDeltaCol] == 0)
                    cross(indiceMaxDeltaCol, false);
                if (tab_deb[indice2][colDemande - 1] == 0)
                    cross(indice2, true);
            }
        }
        return true;
    }

    public void cross(int lOc, boolean crossALine) {
        for (int i = 0; i < (crossALine ? nbC : nbL); i++) {
            if (crossALine)
                crossed[lOc][i] = true;
            else
                crossed[i][lOc] = true;
        }
    }

    private int calculerIndiceMin(int[][] tab, int lOc, boolean deLaLigne) {
        int min = Integer.MAX_VALUE;
        int indiceMin = -1;
        for (int i = 0; i < (deLaLigne ? nbC : nbL); i++) {
            if (deLaLigne) {
                if (!crossed[lOc][i]) {
                    if (tab[lOc][i] <= min) {
                        min = tab[lOc][i];
                        indiceMin = i;
                    }
                }
            } else {
                if (!crossed[i][lOc]) {
                    if (tab[i][lOc] <= min) {
                        min = tab[i][lOc];
                        indiceMin = i;
                    }
                }
            }
        }
        return indiceMin;
    }

    private boolean fini() {
//        for (int i = 0; i < nbL; i++)
//            if (tab_deb[i][colDemande] != 0)
//                return false;
//        for (int i = 0; i < nbC; i++)
//            if (tab_deb[ligStock][i] != 0)
//                return false;
//        return true;

        for (int i = 0; i < nbL; i++)
            for (int j = 0; j < nbC; j++)
                if (!crossed[i][j])
                    return false;
        return true;
    }

    private int calculerDelta(int[][] tab, int lOc, boolean deLaLigne) {
        int min1 = Integer.MAX_VALUE;
        int min2 = Integer.MAX_VALUE;
        for (int i = 0; i < (deLaLigne ? nbC : nbL); i++) {
            if (deLaLigne) {
                if (!crossed[lOc][i]) {
                    if (tab[lOc][i] <= min1)
                        min1 = tab[lOc][i];
                    else if (tab[lOc][i] <= min2)
                        min2 = tab[lOc][i];
                }
            } else {
                if (!crossed[i][lOc]) {
                    if (tab[i][lOc] <= min1)
                        min1 = tab[i][lOc];
                    else if (tab[i][lOc] <= min2)
                        min2 = tab[i][lOc];
                }
            }
        }
        return Math.abs(min1 - min2);
    }

    @Override
    public String toString() {
        String s = "========================================\n";
        for (int i = 0; i < ligStock; i++) {
            for (int j = 0; j < colDemande; j++) {
                s += tab_deb[i][j] + "\t";
            }
            s += "\n";
        }
        s += "----------------------------------------------\n";
        for (int i = 0; i < nbL; i++) {
            for (int j = 0; j < nbC; j++) {
                s += tab_result[i][j] + "\t";
            }
            s += "\n";
        }
        return s;
    }

    public int coutMaximum() {
        int somme = 0;
        for (int i = 0; i < nbL; i++) {
            for (int j = 0; j < nbC; j++) {
                somme += tab_deb[i][j] * tab_result[i][j];
            }
        }
        return somme;
    }
}
